#ifndef lint
static char *RCSid() { return RCSid("$Id: amavis-milter.c,v 1.11 2002/03/02 18:50:24 lhecking Exp $"); }
#endif

/*
 * sendmail/milter client for amavis
 * amavis-perl version
 *
 * Author: Geoff Winkless <gwinkless@users.sourceforge.net>
 * Additional work and patches by:
 *   Gregory Ade
 *   Pierre-Yves Bonnetain
 *   Lars Hecking
 *   Rainer Link
 *   Dale Perkel
 *   Julio Sanchez
 *
 */

/*
 * Add some copyright notice here ...
 */

#ifdef HAVE_CONFIG_H
# include "config.h"
#endif

#define BUFFLEN 255

#include <time.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <sys/stat.h>
#include <sys/utsname.h>
#include <fcntl.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sysexits.h>
#include <unistd.h>
#include <errno.h>
#include <netinet/in.h>
#include <sys/socket.h>
#include <sys/un.h>
#include <stdarg.h>

#ifdef HAVE_SM_GEN_H
# include "sm/gen.h"
#endif
#include "libmilter/mfapi.h"

#ifndef HAVE_SM_GEN_BOOL_TYPE
typedef int bool;
#endif

#ifndef FALSE
# define FALSE	0
#endif /* ! FALSE */
#ifndef TRUE
# define TRUE	1
#endif /* ! TRUE */

#define TEMPLATE "/amavis-milter-XXXXXXXX"
#define DEBUGFILE "/tmp/virdebug"

/* Extracted from the code for better configurability
 * These will be set by configure/make eventually */
#define X_HEADER_TAG "X-Virus-Scanned"
#define X_HEADER_LINE "by amavis-milter (http://amavis.org/)"

typedef struct llstrct {
    char *str;
    struct llstrct *next;
} ll;

struct mlfiPriv {
    struct in_addr client_addr;
    char *mlfi_fname;
    FILE *mlfi_fp;
    char *mlfi_envfrom;
    ll mlfi_envto;
    ll *mlfi_thisenvto;
    int mlfi_numto;
};

#define DBG_NONE    0
#define DBG_INFO    1
#define DBG_WARN    2
#define DBG_FATAL   4

#define DBG_ALL     (DBG_FATAL | DBG_WARN | DBG_INFO)

/* Don't debug by default - use -d N option to switch it on */
static int debuglevel = DBG_NONE;
static const char *mydebugfile = RUNTIME_DIR "/amavis.client";

static void mydebug(const int, const char *, ...);
static char *mymktempdir(char *);
static void freeenvto(ll *);
static sfsistat clearpriv(SMFICTX *, sfsistat, int);
static int allocmem(SMFICTX *);
static sfsistat mlfi_connect(SMFICTX *, char *, _SOCK_ADDR *);
static sfsistat mlfi_envto(SMFICTX *, char **);
static sfsistat mlfi_envfrom(SMFICTX *, char **);
static sfsistat mlfi_header(SMFICTX *, char *, char *);
static sfsistat mlfi_eoh(SMFICTX *);
static sfsistat mlfi_body(SMFICTX *, u_char *, size_t);
static sfsistat mlfi_eom(SMFICTX *);
static sfsistat mlfi_close(SMFICTX *);
static sfsistat mlfi_abort(SMFICTX *);
static sfsistat mlfi_cleanup(SMFICTX *, bool);

static struct utsname myuts;

#if !defined(HAVE_STRLCPY) && !defined(HAVE_SMUTIL_STRLCPY)
# ifdef HAVE_SM_SM_STRLCPY
extern size_t sm_strlcpy (char *, const char *, size_t);
#  define strlcpy sm_strlcpy
# else
static size_t
strlcpy(char *dst, const char *src, size_t size)
{
    size_t src_l = strlen(src);

    if(src_l < size) {
	memcpy(dst, src, src_l + 1);
    } else if(size) {
	memcpy(dst, src, size - 1);
	dst[size - 1] = '\0';
    }
    return src_l;
}
# endif
#endif

static void
mydebug(const int level, const char *fmt, ...)
{
    FILE *f = NULL;
    int fd = -1;
    time_t tmpt;
    char *timestamp;
    va_list ap;

    /* Only bother to do something if we're told to with -d <n> */
    if (!(level & debuglevel))
	return;

    /* Set up debug file */
    if (mydebugfile && (f = fopen(mydebugfile, "a")) == NULL) {
	fprintf(stderr, "error while opening '%s'\n", mydebugfile);
	return;
    } else {
	/* We should never end up here because of the declaration
	 * for mydebugfile above, but if we do.... */
	remove(DEBUGFILE);
	if ((fd = open(DEBUGFILE, O_CREAT | O_EXCL | O_RDWR, 0600)) < 0 || (f = fdopen(fd, "a")) == NULL) {
	    fprintf(stderr, "error while opening '%s'\n", DEBUGFILE);

	    /* don't bother closing an invalid filehandle, but if
	     * it's invalid at *this* point, something is seriously
	     * wrong anyway. */
	    if (fd >= 0)
		close(fd);

	    return;
	}
    }

    tmpt = time(NULL);
    timestamp = ctime(&tmpt);
    /* A 26 character string according ctime(3c)
     * we cut off the trailing \n\0 */
    timestamp[24] = 0;
    fprintf(f, "%s %s amavis(client)[%ld]: ",
	    timestamp, (myuts.sysname ? myuts.sysname : ""), (long) getpid());

    va_start(ap, fmt);
    vfprintf(f, fmt, ap);
    va_end(ap);

    fputc('\n', f);
    fclose(f);
    if (fd >= 0)
	close(fd);
}

static char *
mymktempdir(char *s)
{
    char dirname[BUFFLEN];
    char *stt;
    int count = 0;

    strlcpy(dirname, s, sizeof(dirname));

#ifdef HAVE_MKDTEMP
    return mkdtemp(s);
#else
    /* magic number alert */
    while (count++ < 20) {
# ifdef HAVE_MKTEMP
	stt = mktemp(s);
# else
	stt = strchr(dirname, '-') + 1;
	if (stt) {
	    /* more magic number alert */
	    snprintf(stt, sizeof(dirname) - 1 - (stt - dirname), "%08d", lrand48() / 215);
	    /* This assumes that s in the calling function is the same size
	     * as the local dirname! Beware of malicious coders ;-) */
	    strlcpy(s, dirname, sizeof(dirname));
	    stt = dirname;
	} else {
	    /* invalid template */
	    return NULL;
	}
# endif
	if (stt) {
	    if (!mkdir(s, S_IRWXU)) {
		return s;
	    } else {
		continue;
	    }
	}
    }
    return NULL;
#endif /* HAVE_MKDTEMP */
}

#define MLFIPRIV	((struct mlfiPriv *) smfi_getpriv(ctx))

static void
freeenvto(ll * envto)
{
    ll *new;

    while (envto) {
	new = envto->next;
	if (envto->str) {
	    free(envto->str);
	    envto->str = NULL;
	}
	if (envto)
	    free(envto);
	envto = new;
    }
}

static sfsistat
clearpriv(SMFICTX *ctx, sfsistat retme, int clearall)
{
    /* release private memory and return retme */
    struct mlfiPriv *priv = MLFIPRIV;

    mydebug(DBG_INFO, "Clearing priv (clearall=%d)", clearall);

    if (priv) {
	if (priv->mlfi_fname) {
	    mydebug(DBG_INFO, "clearing fname");
	    free(priv->mlfi_fname);
	    priv->mlfi_fname = NULL;
	}
	if (priv->mlfi_envfrom) {
	    mydebug(DBG_INFO, "clearing envfrom");
	    free(priv->mlfi_envfrom);
	    priv->mlfi_envfrom = NULL;
	}
	if (priv->mlfi_envto.next) {
	    mydebug(DBG_INFO, "clearing multi-envto");
	    freeenvto(priv->mlfi_envto.next);
	    priv->mlfi_envto.next = NULL;
	}
	if (priv->mlfi_envto.str) {
	    mydebug(DBG_INFO, "clearing envto");
	    free(priv->mlfi_envto.str);
	    priv->mlfi_envto.str = NULL;
	}

	mydebug(DBG_INFO, "clearing priv");
	free(priv);
	priv = NULL;
	smfi_setpriv(ctx, priv);
    }
    return retme;
}

/*
 * allocate some private memory if not already allocated
 * returns 0 if ok, 1 if not
 */
static int
allocmem(SMFICTX * ctx)
{
    struct mlfiPriv *priv = MLFIPRIV;

    if (priv == NULL) {
	mydebug(DBG_INFO, "priv was null");
	priv = malloc(sizeof *priv);
	if (priv == NULL) {
	    /* can't accept this message right now */
	    mydebug(DBG_FATAL, "failed to malloc %d bytes for private store: %s",
		    sizeof(*priv), strerror(errno));
	    return 1;
	}
	mydebug(DBG_INFO, "malloced priv - now using memset()");
	memset(priv, 0, sizeof *priv);
	mydebug(DBG_INFO, "malloced priv successfully");
	smfi_setpriv(ctx, priv);
    } else {
	mydebug(DBG_WARN, "allocmem tried but priv was already set");
	mydebug(DBG_WARN, "priv->client_addr.s_addr is %d",
		priv->client_addr.s_addr);
    }
    return 0;
}

static sfsistat
mlfi_connect(SMFICTX * ctx, char *hostname, _SOCK_ADDR * gen_hostaddr)
{
    struct mlfiPriv *priv;
    struct sockaddr_in *hostaddr;

    hostaddr = (struct sockaddr_in *) gen_hostaddr;

    if (hostaddr) {
	mydebug(DBG_INFO, "hostname is %s, addr is %d.%d.%d.%d",
		hostname, (hostaddr->sin_addr.s_addr) & 0xff,
		(hostaddr->sin_addr.s_addr >> 8) & 0xff,
		(hostaddr->sin_addr.s_addr >> 16) & 0xff,
		(hostaddr->sin_addr.s_addr >> 24) & 0xff);
    }
    mydebug(DBG_INFO, "checking allocmem");
    if (allocmem(ctx))
	return SMFIS_TEMPFAIL;
    priv = MLFIPRIV;
    if (hostaddr) {
	priv->client_addr.s_addr = hostaddr->sin_addr.s_addr;
    } else {
	priv->client_addr.s_addr = 0;
    }
    smfi_setpriv(ctx, priv);	/* not really needed but nice */
    return SMFIS_CONTINUE;
}

static sfsistat
mlfi_envto(SMFICTX * ctx, char **envto)
{
    struct mlfiPriv *priv;

    if (allocmem(ctx))
	return SMFIS_TEMPFAIL;

    priv = MLFIPRIV;
    if (!(priv->mlfi_thisenvto)) {
	/* first one... */
	priv->mlfi_thisenvto = &(priv->mlfi_envto);
	priv->mlfi_numto = 1;
    } else {
	priv->mlfi_numto++;
	if ((priv->mlfi_thisenvto->next = malloc(sizeof(ll))) == NULL)
	    return (SMFIS_TEMPFAIL);
	priv->mlfi_thisenvto = priv->mlfi_thisenvto->next;
	priv->mlfi_thisenvto->next = NULL;
    }
    if ((priv->mlfi_thisenvto->str = strdup(*envto)) == NULL)
	return (SMFIS_TEMPFAIL);
    mydebug(DBG_INFO, "added %s as recip", *envto);
    return SMFIS_CONTINUE;
}

static sfsistat
mlfi_envfrom(SMFICTX * ctx, char **envfrom)
{
    struct mlfiPriv *priv;
    char dirname[BUFFLEN];
    char messagecopy[BUFFLEN];
    struct stat buf, StatBuf;

    if (allocmem(ctx))
	return SMFIS_TEMPFAIL;
    priv = MLFIPRIV;

    /* open a file to store this message */
    strlcpy(dirname, RUNTIME_DIR, sizeof(dirname));
    strncat(dirname, TEMPLATE, sizeof(dirname) - 1 - strlen(dirname));
    if (mymktempdir(dirname) == NULL) {
	mydebug(DBG_FATAL, "Failed to create temp dir %s: %s", dirname,
		strerror(errno));
	return SMFIS_TEMPFAIL;
    }
    snprintf(messagecopy, sizeof(messagecopy), "%s/email.txt", dirname);
    mydebug(DBG_INFO, "got %s file, %s sender", messagecopy, *envfrom);

    priv->mlfi_fname = strdup(messagecopy);
    if (priv->mlfi_fname == NULL) {
	return SMFIS_TEMPFAIL;
    }
    priv->mlfi_envfrom = strdup(*envfrom);
    if (!priv->mlfi_envfrom) {
	free(priv->mlfi_fname);
	priv->mlfi_fname = NULL;
	return SMFIS_TEMPFAIL;
    }
    /* umask(0077); */
    if (lstat(dirname, &StatBuf) < 0) {
	mydebug(DBG_FATAL, "Error while trying lstat(%s): %s", dirname,
		strerror(errno));
	exit(SMFIS_TEMPFAIL);
    }

    /* may be too restrictive for you, but's good to avoid problems */
    if (!S_ISDIR(StatBuf.st_mode) || StatBuf.st_uid != geteuid() || StatBuf.st_gid != getegid() || !(StatBuf.st_mode & S_IRWXU)) {
	mydebug(DBG_FATAL,
		"Security Warning: %s must be a Directory and owned by "
		"User %d and Group %d\n"
		"and just read-/write-able by the User and noone else. "
		"Exit.", dirname, geteuid(), getegid());
	exit(EX_TEMPFAIL);
    }
    /* there is still a race condition here if RUNTIME_DIR is writeable by the attacker :-\ */

    if ((priv->mlfi_fp = fopen(priv->mlfi_fname, "w+")) == NULL) {
	free(priv->mlfi_fname);
	priv->mlfi_fname = NULL;
	free(priv->mlfi_envfrom);
	priv->mlfi_envfrom = NULL;
	return SMFIS_TEMPFAIL;
    }

    stat(dirname, &buf);
#if 0
    if (buf.st_uid != geteuid() || (buf.st_mode & 0777) != 0700) {
	mydebug(DBG_FATAL,
		"Security alert -- someone changed %s. Uid is %d, mode is %o",
		dirname, buf.st_uid, 0777 & buf.st_mode);
	fclose(priv->mlfi_fp);
	unlink(priv->mlfi_fname);
	unlink(dirname);
	free(priv->mlfi_fname);
	free(priv->mlfi_envfrom);
	free(priv);
	return SMFIS_TEMPFAIL;
    }
#endif

    /* save the private data */

    smfi_setpriv(ctx, priv);

    /* continue processing */
    return SMFIS_CONTINUE;
}

static sfsistat
mlfi_header(SMFICTX *ctx, char *headerf, char *headerv)
{
    /* write the header to the log file */
    fprintf(MLFIPRIV->mlfi_fp, "%s: %s\n", headerf, headerv);

    /* continue processing */
    return SMFIS_CONTINUE;
}

static sfsistat
mlfi_eoh(SMFICTX *ctx)
{
    /* output the blank line between the header and the body */
    fprintf(MLFIPRIV->mlfi_fp, "\n");

    /* continue processing */
    return SMFIS_CONTINUE;
}

static sfsistat
mlfi_body(SMFICTX *ctx, u_char *bodyp, size_t bodylen)
{
    /* output body block to log file */
    unsigned int y, x;

    /* convert crlf to lf */
    for (x = 0, y = 0; x < bodylen;) {
	bodyp[y++] = bodyp[x++];
	if (bodyp[x] == 13 && bodyp[x + 1] == 10)
	    x++;
    }
    bodylen -= (x - y);

    if (fwrite(bodyp, bodylen, 1, MLFIPRIV->mlfi_fp) <= 0) {
	/* write failed */
	(void) mlfi_cleanup(ctx, FALSE);
	return SMFIS_TEMPFAIL;
    }

    /* continue processing */
    return SMFIS_CONTINUE;
}

static sfsistat
mlfi_eom(SMFICTX *ctx)
{
    struct mlfiPriv *priv = MLFIPRIV;
    int pid, status;

    if (priv) {
	mydebug(DBG_INFO, "EOM");
	/* close the file so we can run checks on it!!! */
	if (priv->mlfi_fp)
	    fclose(priv->mlfi_fp);

	pid = fork();
	if (pid == -1) {
	    mydebug(DBG_FATAL, "couldn't fork: %s", strerror(errno));
	    return clearpriv(ctx, SMFIS_TEMPFAIL, 0);
	}
	if (pid == 0) {
	    char **args;
	    int x;
	    char *tmpstr;

	    args = malloc(sizeof(*args) * (priv->mlfi_numto + 4));
	    tmpstr = malloc(3 * sizeof(*tmpstr) * (priv->mlfi_numto + 3) + 3);
	    if (args == NULL || tmpstr == NULL)
		return clearpriv(ctx, SMFIS_TEMPFAIL, 0);

	    for (x = 0; x < priv->mlfi_numto; x++) {
		strncpy(&(tmpstr[x * 3 + 9]), "%s ", 3);
	    }
	    strncpy(tmpstr, "%s %s %s ", 9);
	    tmpstr[x * 3 + 9] = '\n';
	    tmpstr[x * 3 + 10] = 0;
	    args[0] = "";
	    args[1] = priv->mlfi_fname;
	    args[2] = priv->mlfi_envfrom;
	    priv->mlfi_thisenvto = &(priv->mlfi_envto);
	    for (x = 0; x < priv->mlfi_numto; x++) {
		args[x + 3] = strdup(priv->mlfi_thisenvto->str);
		priv->mlfi_thisenvto = priv->mlfi_thisenvto->next;
	    }

	    args[x + 3] = NULL;
	    strrchr(priv->mlfi_fname, '/')[0] = 0;
	    /* amavis-perl wants the directory, not the filename */
	    mydebug(DBG_INFO, "launching %s %s", tmpstr, args);
	    execv(SBINDIR "/amavis", args);

/* for some reason execl is ignoring the second parameter - so I pass 
   the empty string. Anyone any thoughts as to why this happens??? Geoff */
	    exit(127);
	}
	waitpid(pid, &status, 0);

	if (WIFEXITED(status)) {
	    if (WEXITSTATUS(status) == 127) {
		mydebug(DBG_FATAL, "exec failed, pid %d", pid);
		return clearpriv(ctx, SMFIS_TEMPFAIL, 0);
	    }

	    mydebug(DBG_INFO, "return value was %d\n", WEXITSTATUS(status));

	    if (WEXITSTATUS(status) == 99) {
		mydebug(DBG_WARN, "discarding mail, pid %d", pid);
		return clearpriv(ctx, SMFIS_DISCARD, 0);
		/* there was a virus
		 * discard it so it doesn't go bouncing around!! */
	    }

	    if (WEXITSTATUS(status) == 0) {
		mydebug(DBG_INFO, "adding header");
		smfi_addheader(ctx, X_HEADER_TAG, X_HEADER_LINE);
		mydebug(DBG_WARN, "returning ACCEPT");
		return clearpriv(ctx, SMFIS_ACCEPT, 0);
	    }

	    /* if we got any exit status but 0, we didn't check the file...
	     * so don't add the header. We return TEMPFAIL instead */
	    mydebug(DBG_WARN, "returning TEMPFAIL");
	    return clearpriv(ctx, SMFIS_TEMPFAIL, 0);

	} else {
	    mydebug(DBG_FATAL, "which failed: %s", strerror(errno));
	    return clearpriv(ctx, SMFIS_TEMPFAIL, 0);
	}
    }

    return mlfi_cleanup(ctx, TRUE);

}

static sfsistat
mlfi_close(SMFICTX *ctx)
{
    return clearpriv(ctx, SMFIS_ACCEPT, 1);
}

static sfsistat
mlfi_abort(SMFICTX *ctx)
{
    return mlfi_cleanup(ctx, FALSE);
}

static sfsistat
mlfi_cleanup(SMFICTX *ctx, bool ok)
{
    sfsistat rstat = SMFIS_CONTINUE;
    struct mlfiPriv *priv = MLFIPRIV;
    char *p;
    char host[512];

    if (!priv)
	return rstat;

    /* close the archive file */
    if (priv->mlfi_fp != NULL && fclose(priv->mlfi_fp) == EOF) {
	/* failed; we have to wait until later */
	rstat = SMFIS_TEMPFAIL;
	(void) unlink(priv->mlfi_fname);
    } else if (ok) {
	/* add a header to the message announcing our presence */
	if (gethostname(host, sizeof(host) - 1) < 0)
	    strlcpy(host, "localhost", sizeof host);
	p = strrchr(priv->mlfi_fname, '/');
	if (p == NULL)
	    p = priv->mlfi_fname;
	else
	    p++;
    } else {
	/* message was aborted -- delete the archive file */
	(void) unlink(priv->mlfi_fname);
	*(strrchr(priv->mlfi_fname, '/')) = 0;
	rmdir(priv->mlfi_fname);
    }

    /* return status */
    mydebug(DBG_INFO, "cleanup called");
    return clearpriv(ctx, rstat, 0);
}


struct smfiDesc smfilter = {
    "amavis-milter",		/* filter name */
    SMFI_VERSION,		/* version code -- do not change */
    SMFIF_ADDHDRS,		/* flags */
    mlfi_connect,		/* connection info filter */
    NULL,			/* SMTP HELO command filter */
    mlfi_envfrom,		/* envelope sender filter */
    mlfi_envto,			/* envelope recipient filter */
    mlfi_header,		/* header filter */
    mlfi_eoh,			/* end of header */
    mlfi_body,			/* body block filter */
    mlfi_eom,			/* end of message */
    mlfi_abort,			/* message aborted */
    mlfi_close			/* connection cleanup */
};


int
main(int argc, char *argv[])
{
    int c;
    const char *args = "p:d:";

    umask(0077);

    /* Process command line options */
    while ((c = getopt(argc, argv, args)) != -1) {
	switch (c) {
	case 'p':
	    if (optarg == NULL || *optarg == '\0') {
		mydebug(DBG_FATAL, "Illegal conn: %s", optarg);
		exit(EX_USAGE);
	    }
	    (void) smfi_setconn(optarg);
	    break;
	case 'd':
	    switch (atoi(optarg)) {
	    case 3:
		debuglevel = DBG_INFO;
	    case 2:
		debuglevel = debuglevel | DBG_WARN;
	    case 1:
		debuglevel = debuglevel | DBG_FATAL;
		break;
	    case 0:
		debuglevel = DBG_NONE;
		break;
	    default:
		debuglevel = DBG_ALL;
		break;
	    }
	    break;
	default:
	    break;
	}
    }
    if (smfi_register(smfilter) == MI_FAILURE) {
	mydebug(DBG_FATAL, "smfi_register failed");
	exit(EX_UNAVAILABLE);
    }
    return smfi_main();
}

/* eof */
